/*
 * File: adblock.c
 *
 * Copyright 2005 Rudolf Polzer   <atfield-dt@durchnull.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

/* Adblock implementation.
 * This implementation uses a simple file ~/.dillo/adblock.txt which contains a
 * list of wildcard expressions to be matched. To be compatible with Firefox,
 * lines starting with [ will be ignored.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fnmatch.h>

#include <glib.h>

#include "debug.h"
#include "misc.h"
#include "url.h"

#include "adblock.h"

#define DEBUG_LEVEL 1

#define LINE_MAXLEN 4096
static gboolean disabled;
static GPtrArray *blocked;

static FILE *Adblock_fopen(const char *file, gchar *init_str);

/*
 * Return a file pointer. If the file doesn't exist, try to create it,
 * with the optional 'init_str' as its content.
 */
static FILE *Adblock_fopen(const char *filename, gchar *init_str)
{
   FILE *F_in;
   int fd;

   if ((F_in = fopen(filename, "r+")) == NULL) {
      /* Create the file */
      fd = open(filename, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR);
      if (fd != -1) {
         if (init_str)
            write(fd, init_str, strlen(init_str));
         close(fd);

         DEBUG_MSG(10, "Adblock: Created file: %s\n", filename);
         F_in = Adblock_fopen(filename, NULL);
      } else {
         DEBUG_MSG(10, "Adblock: Could not create file: %s!\n", filename);
      }
   }

   /* set close on exec */
   fcntl(fileno(F_in), F_SETFD, FD_CLOEXEC | fcntl(fileno(F_in), F_GETFD));

   return F_in;
}

void a_Adblock_init()
{
   char *filename;
   char line[LINE_MAXLEN];
   FILE *file_stream;
   guint urls;

   disabled = TRUE;
   
   filename = a_Misc_prepend_user_home(".dillo/adblock.txt");
   file_stream = Adblock_fopen(filename, "[Adblock]\n");
   g_free(filename);

   if (!file_stream) {
      DEBUG_MSG(10, "ERROR: Can't open ~/.dillo/adblock.txt, disabling adblock\n");
      return;
   }

   blocked = g_ptr_array_new();

   urls = 0;
   DEBUG_MSG(10, "Loading Adblock list...\n");
   while (!feof(file_stream)) {
      line[0] = '\0';
      fgets(line, LINE_MAXLEN, file_stream);
      g_strstrip(line);
      if(line[0] == '\0')
         continue;
      if(line[0] == '[')
         continue;
      g_ptr_array_add(blocked, g_strdup(line));
      ++urls;
   }
   DEBUG_MSG(10, "loaded %u adblock entries.\n", urls);

   disabled = FALSE;
}

void a_Adblock_freeall( void )
{
   guint i;
   if(disabled)
      return;
   for(i = 0; i != blocked->len; ++i)
      g_free(g_ptr_array_index(blocked, i));
   g_ptr_array_free(blocked, TRUE);
}

gboolean a_Adblock_isblocked(const DilloUrl *request_url)
{
   guint i;
   DEBUG_MSG(10, "running adblock on %s\n", URL_STR(request_url));
// g_print("running adblock on %s\n", URL_STR(request_url));
   if(disabled)
      return FALSE;
   if(request_url == NULL)
      return FALSE;
   else
   {
      const char *url = URL_STR(request_url);
      for(i = 0; i != blocked->len; ++i)
      {
         const char *pattern = (const char *) g_ptr_array_index(blocked, i);
         if(!fnmatch(pattern, url, 0))
            return TRUE;
      }
      return FALSE;
   }
}

// vim: sw=3 et ts=3
