/*
 * File: i18n.c
 *
 * Copyright (C) 2003 Kiyo <kiyo@teki.jpn.ph>
 * This code referred to Mr. Robert Thomson's code. and created it.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

/*
 * guess encoding charset and convert
 */

#include <config.h>
#include <ctype.h>
#include <string.h>
#include <glib.h>
#include <errno.h>    /* for iconv error codes */
#include <iconv.h>

#include "intl.h"
#include "misc.h"
#include "msg.h"
#include "i18n.h"

//#define DEBUG_LEVEL 10
#include "debug.h"

#define MIN_HEADER_LENGTH (gint)(sizeof("<head><meta http-equiv=\"content-type\"" \
                                       " content=\"text/html\"; charset=></head>") - 1)
#define MAX_HEADER_LENGTH 10240

typedef enum {
   AD_NONE,
   AD_CHINESE,
   AD_KOREAN,
   AD_JAPANESE,
   AD_ALL
} I18nAutoDetectionType;
   
static gchar *default_lang;
static gchar *dw_charset;
static I18nAutoDetectionType detect_type;
static gboolean use_CJK_charset;

typedef struct {
   gchar *lang;
   gchar *charset;
} lang_charset;

#define LANG_TBL_N 59
static lang_charset lang_tbl[LANG_TBL_N] = {
   {"ZH_TW",	"BIG5"},
   {"ZH_HK",	"BIG5HKSCS"},
   {"ZH_CN",	"GBK"},
   {"YI",	"CP1255"},
   {"VI",	"TCVN"},
   {"UR",	"CP1256"},
   {"UK",	"KOI8-U"},
   {"TURKISH",	"ISO8859-9"},
   {"TT",	"TATAR-CYR"},
   {"TR",	"ISO8859-9"},
   {"TG",	"KOI8-C"},
   {"TA",	"TSCII-0"},
   {"SR_SP",	"ISO8859-2"},
   {"SR",	"ISO8859-5"},
   {"SQ",	"ISO8859-2"},
   {"SP",	"ISO8859-5"},
   {"SL",	"ISO8859-2"},
   {"SK",	"ISO8859-2"},
   {"SH",	"ISO8859-2"},
   {"SE_NO",	"UTF-8"},
   {"SERBOCROATIAN",	"ISO8859-2"},
   {"RU_UA",	"KOI8-U"},
   {"RU_RU",	"KOI8-R"},
   {"RUSSIAN",	"ISO8859-5"},
   {"RUMANIAN",	"ISO8859-2"},
   {"RU",	"KOI8-R"},
   {"RO",	"ISO8859-2"},
   {"POSIX",	"C"},
   {"POLISH",	"ISO8859-2"},
   {"PL",	"ISO8859-2"},
   {"MT",	"ISO8859-3"},
   {"MK",	"ISO8859-5"},
   {"LO",	"MULELAO-1"},
   {"KO",	"EUCKR"},
   {"KA",	"GEORGIAN-ACADEMY"},
   {"JP_JP",	"EUCJP"},
   {"JA",	"EUCJP"},
   {"IW",	"ISO8859-8"},
   {"IU",	"NUNACOM-8"},
   {"HY",	"ARMSCII-8"},
   {"HU",	"ISO8859-2"},
   {"HR",	"ISO8859-2"},
   {"HI",	"ISCII-DEV"},
   {"HE",	"ISO8859-8"},
   {"GREEK",	"ISO8859-7"},
   {"FA",	"UTF-8"},
   {"EO",	"ISO8859-3"},
   {"EL",	"ISO8859-7"},
   {"EE",	"ISO8859-4"},
   {"CZ",	"ISO8859-2"},
   {"CS",	"ISO8859-2"},
   {"CROATIAN",	"ISO8859-2"},
   {"CHINESE-T",	"EUCTW"},
   {"CHINESE-S",	"EUCCN"},
   {"BE",	"CP1251"},
   {"AZ",	"ISO8859-9E"},
   {"AR",	"ISO8859-6"},
   {"AM",	"UTF-8"},
   {"A3",	"KOI8-C"},
};

static gchar *I18n_get_charset_from_lang(const gchar *lang)
{
   gint i;
   gchar u_lang[64];

   strncpy(u_lang, lang, 64);
   g_strup(u_lang);
   if (strstr(u_lang, "EURO")) return g_strdup("ISO8859-15");
   for (i = 0; i < LANG_TBL_N; i++) {
      if (strncmp(u_lang, lang_tbl[i].lang, strlen(lang_tbl[i].lang)) == 0)
         return g_strdup(lang_tbl[i].charset);
   }
   return g_strdup("ISO8859-1");
}

void a_I18n_detect_DW_locale()
{
   char *lang;

   lang = getenv("LC_CTYPE");
   if (!lang || !*lang) lang = getenv("LC_ALL");
   if (!lang || !*lang) lang = getenv("LANG");

   detect_type = AD_NONE;
   use_CJK_charset = FALSE;
   if (!lang) {
      dw_charset = g_strdup("ISO8859-1");
      default_lang = g_strdup("en");
      return;
   }
   if (g_strncasecmp(lang, "ZH", 2) == 0) {
      detect_type = AD_CHINESE;
      use_CJK_charset = TRUE; 
   } else if (g_strncasecmp(lang, "KO", 2) == 0) {
      detect_type = AD_KOREAN;
      use_CJK_charset = TRUE; 
   } else if (g_strncasecmp(lang, "JA", 2) == 0) {
      detect_type = AD_JAPANESE;
      use_CJK_charset = TRUE; 
   }
   if (strchr(lang, '.')) {
      default_lang = g_strdup(lang);
      strchr(default_lang, '.')[0] = 0;
   } else {
      default_lang = g_strdup(lang);
   }
   if (strchr(default_lang, '_')) {
      strchr(default_lang, '_')[0] = '-';
   }
   
   dw_charset = (strchr(lang, '.'))
           ? a_I18n_fix_charset(g_strdup(strchr(lang, '.') + 1))
           : I18n_get_charset_from_lang(lang);
   if (!dw_charset) dw_charset = g_strdup(DILLO_CHARSET);
}

const gchar *a_I18n_get_DW_charset()
{
   return dw_charset;
}

/* If charset has latin1, return TRUE. */
gboolean a_I18n_has_latin1_charset(gchar *lang)
{
   return (g_strncasecmp(lang, "ja", 2) != 0);
}

gchar *a_I18n_get_lang_from_charset(const gchar *charset)
{
   gchar *buf, *lang;
   if (!charset) return NULL;

   buf = g_strdup(charset);
   g_strup(buf);
   if (strcmp(buf, "ISO-8859-1") == 0 ||
         strcmp(buf, "WINDOWS-1252") == 0) {
      lang ="en"; /* af,sq,eu,ca,da,nl,fo,fi,fr,gl,de,is,ga,it */
   } else if (strncmp(buf, "BIG", 3) == 0 ||
         strcmp(buf, "EUC-TW") == 0) { 
      lang ="zh-TW"; /* Traditional */
   } else if (strncmp(buf, "GB", 2) == 0 ||
         strcmp(buf, "EUC-CN") == 0) {
      lang ="zh-CN"; /* Simplified */
   } else if (strcmp(buf, "SJIS") == 0 ||
         strncmp(buf, "SHIFT", 5) == 0 ||
         strcmp(buf, "EUC-JP") == 0 ||
         strcmp(buf, "ISO-2022-JP") == 0) {
      lang ="ja";
   } else if (strcmp(buf, "EUC-KR") == 0 ||
         strcmp(buf, "ISO-2022-KR") == 0) {
      lang ="ko";
   } else if (strcmp(buf, "ISO-8859-2") == 0 ||
         strcmp(buf, "WINDOWS-1250") == 0) {
      lang ="hr"; /* sl */
   } else if (strcmp(buf, "ISO-8859-3") == 0) {
      lang ="eo"; /* mt */
   } else if (strcmp(buf, "WINDOWS-874") == 0) {
      lang ="th";
   } else if (strcmp(buf, "WINDOWS-1251") == 0 ||
         strcmp(buf, "KOI8-R") == 0) {
      lang ="ru";
   } else if (strcmp(buf, "ISO-8859-5") == 0 ||
         strcmp(buf, "KOI8-U") == 0) {
      lang ="uk";
   } else if (strcmp(buf, "ISO-8859-6") == 0 ||
         strcmp(buf, "WINDOWS-1256") == 0) {
      lang ="ar";
   } else if (strcmp(buf, "ISO-8859-7") == 0) {
      lang ="el";
   } else if (strcmp(buf, "ISO-8859-8") == 0) {
      lang ="he";
   } else if (strcmp(buf, "ISO-8859-9") == 0 ||
         strcmp(buf, "WINDOWS-1254") == 0) {
      lang ="tr";
   } else if (strcmp(buf, "ISO-8859-13") == 0 ||
         strcmp(buf, "WINDOWS-1257") == 0) {
      lang ="lt"; /* lv */
   } else if (strcmp(buf, "ISO-8859-15") == 0) {
      lang = "et";
/* eo es la ms ...
   } else if (strncmp(buf, "", 1) == 0) {
      lang = "";
*/
   } else {
      lang = default_lang;
   }
   g_free(buf);

   return g_strdup(lang);;
}

gboolean a_I18n_use_CJK_charset(gchar *lang)
{
   return (g_strncasecmp(lang, "zh", 2) == 0 ||
         g_strncasecmp(lang, "ko", 2) == 0 ||
         g_strncasecmp(lang, "ja", 2) == 0);
}

/* fix name Html_charset to iconv_charset. for example "x-sjis" to "SJIS"... */
/* TODO: remake this code! */
gchar *a_I18n_fix_charset(gchar *charset)
{
   iconv_t it;
   
   if (!charset) return NULL;
   g_strup(charset);
   if (strchr(charset, '@'))
      strchr(charset, '@')[0] = '\0';
   if (strncmp(charset, "X-", 2) == 0) {
      gchar *old_charset = charset;
      charset = g_strdup(charset + 2);
      g_free(old_charset);
   }
   /* Japan Only */
   if (strcmp(charset, "SHIFTJIS") == 0) {
      g_free(charset);
      return g_strdup("SJIS");
   }
   /* Korea Only */
   if (strcmp(charset, "KS_C_5601-1987") == 0) {
      g_free(charset);
      return g_strdup("EUC-KR");
   }
   if (strcmp(charset, "UNICODE-1-1-UTF-8") == 0) {
      g_free(charset);
      return g_strdup("UTF-8");
   }
   if (strchr(charset, '_')) {
      strchr(charset, '_')[0] = '-';
   }
   if ((it = iconv_open(charset, charset)) == (iconv_t)-1) {
      g_free(charset);
      return NULL;
   }
   iconv_close(it);
   return charset;
}

/* auto detect Chinese charset. -- GB2312,GBK,GB18030,HZ,ISO-2022-CN,
 * Big5,Big5-HKSCS,EUC-CN,EUC-TW*/
gchar *I18n_detect_chinese_charset(const gchar *buf, gint bufsize)
{
   /* TODO: implement this! */
   return g_strdup(DW_CHARSET);
}

/* auto detect Korean charset. -- EUC-KR,UHC,JOHAB,ISO-2022-KR */
gchar *I18n_detect_korean_charset(const gchar *buf, gint bufsize)
{
   /* TODO: implement this! */
   return g_strdup(DW_CHARSET);
}

/* auto detect Japanese charset. -- Shift_JIS,EUC-JP,ISO-2022-JP */
gchar *I18n_detect_japanese_charset(const gchar *buf, gint bufsize)
{
   gint i, got_sjis = 0, got_euc = 0, got_jis = 0;
   if (!buf || bufsize == 0)
      return g_strdup("SJIS");
   
   /* check CR code. */
   if (strstr(buf, "\r"))
      got_sjis += 2;
         
   for (i = 0; i < bufsize; i++) {
      if (got_sjis > got_euc && got_sjis >= 40)
         return g_strdup("SJIS");
      else if (got_euc > got_sjis && got_euc >= 40)
         return g_strdup("EUC-JP");
      else if (got_jis > 3)
         return g_strdup("ISO-2022-JP");
      if (buf[i] == 0x1b) {
         if (buf[i+1] == '$' || buf[i+1] == '(')
            got_jis++;
         else
            got_jis--;
      } else if (IS_ASCII_CHAR(buf[i])) continue;
      _MSG("e=%d s=%d j=%d %d\n", got_euc, got_sjis, got_jis, i);
      if (IS_SJIS_CHAR1(buf[i])) {
         if(IS_SJIS_CHAR2(buf[i+1])) {
            i++;
            got_sjis += 2;
            continue;
         } else
            got_euc += 2;
      }
      if (IS_EUC_CHAR(buf[i])) {
         if (IS_EUC_CHAR(buf[i+1])) {
            got_euc += 2;
            i++;
         } else got_sjis += 2;
      }
      if (i >= 1)
         if (IS_HANKANA_CHAR(buf[i - 2])
               && IS_HANKANA_CHAR(buf[i - 1])
               && IS_HANKANA_CHAR(buf[i])) {
            got_sjis++;
            if (IS_HANKANA_CHAR(buf[i+1]))
               got_sjis++;
      }
   }

   _MSG(_("charset detect may mistake...\n"));
   if (got_jis > 1)
      return g_strdup("ISO-2022-JP");
   if (got_sjis > got_euc)
      return g_strdup("SJIS");
   if (got_euc >= 1)
      return g_strdup("EUC-JP");
   
   /* check CR code again... */
   if (strstr(buf, "\r"))
      return g_strdup("SJIS");
   else
      return g_strdup("EUC-JP");
}

static gchar *I18n_detect_utf8(const gchar *buf, gint bufsize)
{
   gint i, utf_count;

   utf_count = 0;
   for (i = 0; i < bufsize; i++) {
      if (IS_UTF8_FIRST_CHAR(buf[i])) {
         if (IS_ASCII_CHAR(buf[i + 1])) {
            utf_count -= 8;
         } else if (IS_UTF8_2BYTE_FIRST_CHAR(buf[i])) {
            if (IS_UTF8_NON_FIRST_CHAR(buf[i + 1])
                  && (IS_UTF8_FIRST_CHAR(buf[i + 2])
                     || IS_ASCII_CHAR(buf[i + 2])))
               utf_count++;
            else
               utf_count -= 8;
         } else if (IS_UTF8_3BYTE_FIRST_CHAR(buf[i])) {
            if (IS_UTF8_NON_FIRST_CHAR(buf[i + 1])
                  && IS_UTF8_NON_FIRST_CHAR(buf[i + 2])
                  && (IS_UTF8_FIRST_CHAR(buf[i + 3])
                     || IS_ASCII_CHAR(buf[i + 3])))
               utf_count += 2;
            else
               utf_count -= 8;
         } else if (IS_UTF8_4BYTE_FIRST_CHAR(buf[i])
               && IS_UTF8_NON_FIRST_CHAR(buf[i + 1])
               && IS_UTF8_NON_FIRST_CHAR(buf[i + 2])
               && IS_UTF8_NON_FIRST_CHAR(buf[i + 3])
               && (IS_UTF8_FIRST_CHAR(buf[i + 4])
                  || IS_ASCII_CHAR(buf[i + 4])))
            utf_count += 3;
         else
            utf_count -= 8;
         if (utf_count <= -16 || utf_count >= 16)
            break;
      }
   }
   if (utf_count >= 9)
      return g_strdup("UTF-8");
   return NULL;
}

gchar *a_I18n_get_charset_from_meta_tag(
      const gchar *buf, gint bufsize)
{
   gchar *meta_buf, *meta_tag, *head_end, *meta_charset, *charset = NULL;
   gint meta_buf_size, charset_len;

   if (!buf || bufsize < MIN_HEADER_LENGTH)
      return NULL;
   
   /* crate upper string */
   if (bufsize > MAX_HEADER_LENGTH)
      meta_buf_size = MAX_HEADER_LENGTH;
   else
      meta_buf_size = bufsize;
   meta_buf = g_strndup(buf, meta_buf_size);
   meta_buf[meta_buf_size - 1] = '\0';
   g_strup(meta_buf);
   
   /* find meta-charset */
   if (!((meta_tag = strstr(meta_buf, "<META "))
      && (meta_tag += sizeof("<META ") - 1)
      && (meta_charset = strstr(meta_tag, "CHARSET="))
      && (strstr(meta_tag, "HTTP-EQUIV=\"CONTENT-TYPE")
         || strstr(meta_tag, "HTTP-EQUIV=CONTENT-TYPE")))) {
      g_free(meta_buf);
      return NULL;
   }
   head_end = strstr(meta_tag, "</HEAD>");

   /* check HTML code. */
   if (meta_charset > head_end) {
      MSG(_("<meta> outside of <head>!\n"));
      //g_free(meta_buf);
      //return NULL;
   }

   meta_charset += sizeof("CHARSET=") - 1;
   if (meta_charset[0] == '"' || meta_charset[0] == '\'')
      meta_charset++;
   charset_len = strpbrk(meta_charset, "\"'/> ") - meta_charset;
   if (charset_len > 0)
      charset = g_strndup(meta_charset, charset_len);
   g_free(meta_buf);
   if (charset)
      charset = a_I18n_fix_charset(charset);
   if (charset && strncmp(charset, "UTF", 3) == 0
         && (strncmp(charset, "UTF16", 5) == 0
         || strncmp(charset, "UTF-16", 6) == 0
         || strncmp(charset,  "UTF32", 5) == 0
         || strncmp(charset, "UTF-32", 6) == 0))
      return NULL;
   return charset;
}

/** get encoding charset.
 * buf     : check target buffer.
 * bufsize : buf's size.
 * return  : charset
 */
gchar *a_I18n_get_encoding(const gchar *buf, gint bufsize)
{
   gchar *charset;

   if (!buf) return NULL;
   if (bufsize == 0) return g_strdup("");

   /* skip ascii char */
   {
      gint i, old_bufsize = bufsize;
      for (i = 0; i < old_bufsize; i++) {
         if (IS_ASCII_CHAR(buf[0])
               && buf[0] != 0x1b) { //for ISO-2022-* detect
            buf++;
            bufsize--;
         } else break;
      }
      if (bufsize == 0) {
         _MSG(_("all ascii code\n"));
         return g_strdup("ASCII");
      }
   }

   /* Check UTF-8 code. */
   charset = I18n_detect_utf8(buf, bufsize);
   if (charset) return charset;

   switch (detect_type) {
      case AD_ALL:
         /* TODO: implement this! */
      case AD_CHINESE:
         return I18n_detect_chinese_charset(buf, bufsize);
      case AD_KOREAN:
         return I18n_detect_korean_charset(buf, bufsize);
      case AD_JAPANESE:
         return I18n_detect_japanese_charset(buf, bufsize);
      case AD_NONE:
      default:
         return g_strdup(DW_CHARSET);
   }
}

/* [hack] iconv with error recovery */
size_t a_I18n_iconv(iconv_t cd,
      char **inbuf, size_t *inbytesleft,
      char **outbuf, size_t *outbytesleft)
{
   size_t error_count = 0;
   
   if (iconv(cd, inbuf, inbytesleft, outbuf, outbytesleft)
       == (size_t)-1 && errno != EINVAL) {
      _MSG(_("unable to fully convert between character encodings\n"));
      /*
       * This just skips past unconvertable characters, putting white square
       * in the output, then retries the conversion.
       */
      while ((int)*inbytesleft > 0 && (int)*outbytesleft > 0
            && errno == EILSEQ) {
         error_count++;
         (*outbuf)[0]   = '['; /* white square */
         (*outbuf)[1]   = ']';
         *outbuf       += 2;
         *outbytesleft -= 2;
         *inbuf        += 2;
         *inbytesleft  -= 2;
         if ((int)*inbytesleft > 0 && (int)*outbytesleft > 0)
            iconv(cd, inbuf, inbytesleft, outbuf, outbytesleft);
      }
   }
   *outbuf[0] = '\0';
   return error_count;
}


/* Charset Transrate. for the part which gives priority to speed.*/
gchar *a_I18n_convert_raw(iconv_t *it, const gchar* buf, size_t bufsize)
{
   size_t outLeft, outSize;
   gchar *outStart, *in, *out;

   outLeft = bufsize * D_CHAR_GROWTH;
   outSize = outLeft + 1;
   if (!buf) return NULL;
   if (bufsize == 0) return g_strdup("");

   outStart = (gchar *)g_malloc(outSize);
   in  = (gchar *)buf;
   out = outStart;

   if (!*it || *it==(iconv_t)-1) {
      g_warning(_("no support enc\n"));
      *it = iconv_open("ASCII", DILLO_CHARSET);
   }
   a_I18n_iconv(*it, &in, &bufsize, &out, &outLeft);
   _MSG("free %d bytes.\n", outSize - outLeft);
   return outStart = g_realloc(outStart, outSize - outLeft);
}

/* Charset Transrate. for simple text or string.
 * if buf is nul-terminated, can set bufsize -1 */
gchar *a_I18n_convert(
      const gchar* from, const gchar* to,
      const gchar* buf, gint bufsize)
{
   iconv_t it;
   gchar *outBuf;
   gint len;

   if (!buf) return NULL;
   if (bufsize == 0) return g_strdup("");
   g_return_val_if_fail(from && to, g_strndup(buf, bufsize));
   len = (bufsize == -1) ? (gint)strlen(buf) : bufsize;
   if (len < 0) {
      MSG("a_I18n_convert: bufsize < 0\n");
      return g_strdup(buf);
   }
   it = iconv_open(to, from);
   outBuf = a_I18n_convert_raw(&it, buf, len);
   iconv_close(it);
   return outBuf;
}

/* Convert to Toolkit's charset. */
gchar *a_I18n_DW_CHARSET_from_DILLO_CHARSET(const gchar* buf, gint bufsize)
{
   return a_I18n_convert(DILLO_CHARSET, DW_CHARSET, buf, bufsize);
}

/* Convert to internal charset. */
gchar *a_I18n_DW_CHARSET_to_DILLO_CHARSET(const gchar* buf, gint bufsize)
{
   return a_I18n_convert(DW_CHARSET, DILLO_CHARSET, buf, bufsize);
}

/* For gettext string. */
gchar *a_I18n_locale_to_DILLO_CHARSET(const gchar* buf)
{
   return a_I18n_convert(LOCALE_CHARSET, DILLO_CHARSET, buf, -1);
}

/* mblen() only for UTF-8 stream.
 * return multibyte char size. */
gint a_I18n_mblen(gchar *str)
{
   if (IS_UTF8_3BYTE_FIRST_CHAR(str[0])) {
      if (IS_UTF8_NON_FIRST_CHAR(str[1]))
         if (IS_UTF8_NON_FIRST_CHAR(str[2]))
            return 3;
   } else if(IS_UTF8_2BYTE_FIRST_CHAR(str[0])) {
      if (IS_UTF8_NON_FIRST_CHAR(str[1]))
         return 2;
   } else if(IS_UTF8_4BYTE_FIRST_CHAR(str[0])) {
      if (IS_UTF8_NON_FIRST_CHAR(str[1]))
         if (IS_UTF8_NON_FIRST_CHAR(str[2]))
            if (IS_UTF8_NON_FIRST_CHAR(str[3]))
               return 4;
   }
   return 0;
}

/*
 * check invalid string, and repair.
 */
gchar *a_I18n_string_check(const gchar *input, const gchar *charset)
{
   gint i, len;
   gchar *ret, *str;
   GString *out = g_string_new("");
   
   /* convert to same charset.
    * invalid chars will be replaced to white square. */
   str = a_I18n_convert(charset, charset, input, -1);
   len = (str) ? strlen(str) : 0;

   /* skip white square. */
   for(i = 0; i < len; i++) {
      if ((guchar)str[i] == '[' && (guchar)str[i+1] == ']')
         i++;
      else
         g_string_append_c(out, str[i]);
   }
   g_free(str);
   ret = out->str;
   g_string_free(out, FALSE);
   return ret;
}

/* vim: set ts=3 sw=3 sts=3 expandtab:*/
